@@ -56,7 +56,7 @@ libvisio::VSDXContentCollector::VSDXContentCollector(
   m_painter(painter), m_isPageStarted(false), m_pageWidth(0.0), m_pageHeight(0.0),
   m_shadowOffsetX(0.0), m_shadowOffsetY(0.0),
   m_scale(1.0), m_x(0.0), m_y(0.0), m_originalX(0.0), m_originalY(0.0), m_xform(),
-  m_txtxform(0), m_currentGeometry(), m_groupXForms(groupXFormsSequence[0]),
+  m_txtxform(0), m_currentFillGeometry(), m_currentLineGeometry(), m_groupXForms(groupXFormsSequence[0]),
   m_currentForeignData(), m_currentOLEData(), m_currentForeignProps(),
   m_currentShapeId(0), m_foreignType(0), m_foreignFormat(0), m_foreignOffsetX(0.0),
   m_foreignOffsetY(0.0), m_foreignWidth(0.0), m_foreignHeight(0.0), m_styleProps(),
@@ -528,21 +528,21 @@ const char *libvisio::VSDXContentCollector::_linePropertiesMarkerPath(unsigned m
   case 13:
     return "m10 0-10 30h20z";
   case 14:
-    return "m10 0-10 30h20z m-10 -18l-5 15h10z";
+    return "m10 0-10 30h20z m0 12l-5 15h10z";
   case 15:
-    return "m10 0-10 10h20z m-10 -7l-5 5h10z";
+    return "m10 0-10 10h20z m0 3l-5 5h10z";
   case 16:
-    return "m10 0-10 20h20z m-10 -13l-5 10h10z";
+    return "m10 0-10 20h20z m0 7l-5 10h10z";
   case 17:
-    return "m10 0-10 20q10,-5 20,0z m-10 -13l-4 8q4,-2 8,0z";
+    return "m10 0-10 20q10,-5 20,0z m0 7l-4 8q4,-2 8,0z";
   case 18:
-    return "m10 0-10 20q10,5 20,0z m-10 -13l-5 10q5,2 10,0z";
+    return "m10 0-10 20q10,5 20,0z m0 7l-5 10q5,2 10,0z";
   case 21:
-    return "m0 0v30h30v-30z m-20 10v10h10v-10z";
+    return "m0 0v30h30v-30z m10 10v10h10v-10z";
   case 22:
-    return "m10 0-10 20l10 20l10 -20z m-10 -12l-6 12l6 12l6 -12z";
+    return "m10 0-10 20l10 20l10 -20z m0 8l-6 12l6 12l6 -12z";
   case 39:
-    return "m10 0-10 20h20z m-10 0-10 20h20z";
+    return "m10 0-10 20h20z m0 20-10 20h20z";
   default:
     return "m10 0-10 30h20z";
   }
@@ -570,64 +570,78 @@ double libvisio::VSDXContentCollector::_linePropertiesMarkerScale(unsigned marke
 void libvisio::VSDXContentCollector::_flushCurrentPath()
 {
   WPXPropertyListVector path;
-  double startX = 0;
-  double startY = 0;
-  double x = 0;
-  double y = 0;
+  WPXPropertyList fillPathProps(m_styleProps);
+  fillPathProps.insert("draw:stroke", "none");
+  WPXPropertyList linePathProps(m_styleProps);
+  linePathProps.insert("draw:fill", "none");
   bool firstPoint = true;
   bool wasMove = false;
+  bool needsGroup = true;
 
-  for (unsigned i = 0; i < m_currentGeometry.size(); i++)
+  if (!m_styleProps["draw:fill"] || m_styleProps["draw:fill"]->getStr() == "none")
+    needsGroup = false;
+  if (m_currentFillGeometry.empty())
+    needsGroup = false;
+  if (!m_styleProps["draw:stroke"] || m_styleProps["draw:stroke"]->getStr() == "none")
+    needsGroup = false;
+  if (m_currentLineGeometry.empty())
+    needsGroup = false;
+
+  if (needsGroup)
+    m_shapeOutputDrawing->addStartLayer(WPXPropertyList());
+
+  if (m_styleProps["draw:fill"] && m_styleProps["draw:fill"]->getStr() != "none")
   {
-    if (firstPoint)
+    for (unsigned i = 0; i < m_currentFillGeometry.size(); i++)
     {
-      x = m_currentGeometry[i]["svg:x"]->getDouble();
-      y = m_currentGeometry[i]["svg:y"]->getDouble();
-      startX = x;
-      startY = y;
-      firstPoint = false;
-      wasMove = true;
-    }
-    else if (m_currentGeometry[i]["libwpg:path-action"]->getStr() == "M")
-    {
-      if (((startX == x && startY == y)
-           || (m_styleProps["draw:fill"] && m_styleProps["draw:fill"]->getStr() != "none"
-               && m_styleProps["draw:stroke"] && m_styleProps["draw:stroke"]->getStr() == "none"))
-          && path.count() && !wasMove)
+      if (firstPoint)
       {
-        WPXPropertyList closedPath;
-        closedPath.insert("libwpg:path-action", "Z");
-        path.append(closedPath);
+        firstPoint = false;
+        wasMove = true;
       }
-      x = m_currentGeometry[i]["svg:x"]->getDouble();
-      y = m_currentGeometry[i]["svg:y"]->getDouble();
-      startX = x;
-      startY = y;
-      wasMove = true;
+      else if (m_currentFillGeometry[i]["libwpg:path-action"]->getStr() == "M")
+      {
+        if (path.count() && !wasMove)
+        {
+          WPXPropertyList closedPath;
+          closedPath.insert("libwpg:path-action", "Z");
+          path.append(closedPath);
+        }
+        wasMove = true;
+      }
+      else
+        wasMove = false;
+      path.append(m_currentFillGeometry[i]);
     }
-    else
+    if (path.count() && !wasMove)
     {
-      x = m_currentGeometry[i]["svg:x"]->getDouble();
-      y = m_currentGeometry[i]["svg:y"]->getDouble();
-      wasMove = false;
+      WPXPropertyList closedPath;
+      closedPath.insert("libwpg:path-action", "Z");
+      path.append(closedPath);
+    }
+    if (path.count())
+    {
+      m_shapeOutputDrawing->addStyle(fillPathProps, WPXPropertyListVector());
+      m_shapeOutputDrawing->addPath(path);
     }
-    path.append(m_currentGeometry[i]);
   }
-  if (((startX == x && startY == y)
-       || (m_styleProps["draw:fill"] && m_styleProps["draw:fill"]->getStr() != "none"
-           && m_styleProps["draw:stroke"] && m_styleProps["draw:stroke"]->getStr() == "none"))
-      && path.count() && !wasMove)
+  m_currentFillGeometry.clear();
+  path = WPXPropertyListVector();
+
+  if (m_styleProps["draw:stroke"] && m_styleProps["draw:stroke"]->getStr() != "none")
   {
-    WPXPropertyList closedPath;
-    closedPath.insert("libwpg:path-action", "Z");
-    path.append(closedPath);
+    for (unsigned i = 0; i < m_currentLineGeometry.size(); i++)
+      path.append(m_currentLineGeometry[i]);
+    if (path.count())
+    {
+      m_shapeOutputDrawing->addStyle(linePathProps, WPXPropertyListVector());
+      m_shapeOutputDrawing->addPath(path);
+    }
   }
-  if (path.count() && !m_noShow)
-  {
-    m_shapeOutputDrawing->addStyle(m_styleProps, WPXPropertyListVector());
-    m_shapeOutputDrawing->addPath(path);
-  }
-  m_currentGeometry.clear();
+  m_currentLineGeometry.clear();
+
+  if (needsGroup)
+    m_shapeOutputDrawing->addEndLayer();
 }
 
 void libvisio::VSDXContentCollector::_flushText()
@@ -968,7 +982,10 @@ void libvisio::VSDXContentCollector::collectEllipticalArcTo(unsigned /* id */, u
     end.insert("svg:x", m_scale*m_x);
     end.insert("svg:y", m_scale*m_y);
     end.insert("libwpg:path-action", "L");
-    m_currentGeometry.push_back(end);
+    if (!m_noFill && !m_noShow)
+      m_currentFillGeometry.push_back(end);
+    if (!m_noLine && !m_noShow)
+      m_currentLineGeometry.push_back(end);
     return;
   }
 
@@ -1005,7 +1022,10 @@ void libvisio::VSDXContentCollector::collectEllipticalArcTo(unsigned /* id */, u
   arc.insert("svg:x", m_scale*m_x);
   arc.insert("svg:y", m_scale*m_y);
   arc.insert("libwpg:path-action", "A");
-  m_currentGeometry.push_back(arc);
+  if (!m_noFill && !m_noShow)
+    m_currentFillGeometry.push_back(arc);
+  if (!m_noLine && !m_noShow)
+    m_currentLineGeometry.push_back(arc);
 }
 
 void libvisio::VSDXContentCollector::collectEllipse(unsigned /* id */, unsigned level, double cx, double cy, double xleft, double yleft, double xtop, double ytop)
@@ -1030,7 +1050,10 @@ void libvisio::VSDXContentCollector::collectEllipse(unsigned /* id */, unsigned
   ellipse.insert("svg:x",m_scale*xleft);
   ellipse.insert("svg:y",m_scale*yleft);
   ellipse.insert("libwpg:path-action", "M");
-  m_currentGeometry.push_back(ellipse);
+  if (!m_noFill && !m_noShow)
+    m_currentFillGeometry.push_back(ellipse);
+  if (!m_noLine && !m_noShow)
+    m_currentLineGeometry.push_back(ellipse);
   ellipse.insert("svg:rx",m_scale*rx);
   ellipse.insert("svg:ry",m_scale*ry);
   ellipse.insert("svg:x",m_scale*xtop);
@@ -1038,12 +1061,17 @@ void libvisio::VSDXContentCollector::collectEllipse(unsigned /* id */, unsigned
   ellipse.insert("libwpg:large-arc", largeArc?1:0);
   ellipse.insert("libwpg:path-action", "A");
   ellipse.insert("libwpg:rotate", angle * 180/M_PI, WPX_GENERIC);
-  m_currentGeometry.push_back(ellipse);
+  if (!m_noFill && !m_noShow)
+    m_currentFillGeometry.push_back(ellipse);
+  if (!m_noLine && !m_noShow)
+    m_currentLineGeometry.push_back(ellipse);
   ellipse.insert("svg:x",m_scale*xleft);
   ellipse.insert("svg:y",m_scale*yleft);
   ellipse.insert("libwpg:large-arc", largeArc?0:1);
-  m_currentGeometry.push_back(ellipse);
-
+  if (!m_noFill && !m_noShow)
+    m_currentFillGeometry.push_back(ellipse);
+  if (!m_noLine && !m_noShow)
+    m_currentLineGeometry.push_back(ellipse);
 }
 
 void libvisio::VSDXContentCollector::collectInfiniteLine(unsigned /* id */, unsigned level, double x1, double y1, double x2, double y2)
@@ -1119,11 +1147,17 @@ void libvisio::VSDXContentCollector::collectInfiniteLine(unsigned /* id */, unsi
   infLine.insert("svg:x",m_scale*xmove);
   infLine.insert("svg:y",m_scale*ymove);
   infLine.insert("libwpg:path-action", "M");
-  m_currentGeometry.push_back(infLine);
+  if (!m_noFill && !m_noShow)
+    m_currentFillGeometry.push_back(infLine);
+  if (!m_noLine && !m_noShow)
+    m_currentLineGeometry.push_back(infLine);
   infLine.insert("svg:x",m_scale*xline);
   infLine.insert("svg:y",m_scale*yline);
   infLine.insert("libwpg:path-action", "L");
-  m_currentGeometry.push_back(infLine);
+  if (!m_noFill && !m_noShow)
+    m_currentFillGeometry.push_back(infLine);
+  if (!m_noLine && !m_noShow)
+    m_currentLineGeometry.push_back(infLine);
 }
 
 void libvisio::VSDXContentCollector::collectLine(unsigned /* id */, unsigned level, double strokeWidth, Colour c, unsigned linePattern, unsigned char startMarker, unsigned char endMarker, unsigned lineCap)
@@ -1278,24 +1312,17 @@ void libvisio::VSDXContentCollector::collectGeometry(unsigned /* id */, unsigned
   m_y = 0.0;
   m_originalX = 0.0;
   m_originalY = 0.0;
-  bool noFill = ((geomFlags & 1) == 1);
-  bool noLine = ((geomFlags & 2) == 2);
-  bool noShow = ((geomFlags & 4) == 4);
-
-  if ((m_noFill != noFill) || (m_noLine != noLine) || (m_noShow != noShow) || m_isFirstGeometry)
-    _flushCurrentPath();
+  m_noFill = ((geomFlags & 1) == 1);
+  m_noLine = ((geomFlags & 2) == 2);
+  m_noShow = ((geomFlags & 4) == 4);
 
   _applyLinePattern();
 
-  m_isFirstGeometry = false;
-  m_noFill = noFill;
-  m_noLine = noLine;
-  m_noShow = noShow;
-  if (m_noLine || m_linePattern == 0)
+  if (m_linePattern == 0)
     m_styleProps.insert("draw:stroke", "none");
   else
     m_styleProps.insert("svg:stroke-color", m_lineColour);
-  if (m_noFill || m_fillPattern == 0)
+  if (m_fillPattern == 0)
     m_styleProps.insert("draw:fill", "none");
   else
   {
@@ -1318,7 +1345,10 @@ void libvisio::VSDXContentCollector::collectMoveTo(unsigned /* id */, unsigned l
   end.insert("svg:x", m_scale*m_x);
   end.insert("svg:y", m_scale*m_y);
   end.insert("libwpg:path-action", "M");
-  m_currentGeometry.push_back(end);
+  if (!m_noFill && !m_noShow)
+    m_currentFillGeometry.push_back(end);
+  if (!m_noLine && !m_noShow)
+    m_currentLineGeometry.push_back(end);
 }
 
 void libvisio::VSDXContentCollector::collectLineTo(unsigned /* id */, unsigned level, double x, double y)
@@ -1333,7 +1363,10 @@ void libvisio::VSDXContentCollector::collectLineTo(unsigned /* id */, unsigned l
   end.insert("svg:x", m_scale*m_x);
   end.insert("svg:y", m_scale*m_y);
   end.insert("libwpg:path-action", "L");
-  m_currentGeometry.push_back(end);
+  if (!m_noFill && !m_noShow)
+    m_currentFillGeometry.push_back(end);
+  if (!m_noLine && !m_noShow)
+    m_currentLineGeometry.push_back(end);
 }
 
 void libvisio::VSDXContentCollector::collectArcTo(unsigned /* id */, unsigned level, double x2, double y2, double bow)
@@ -1353,7 +1386,10 @@ void libvisio::VSDXContentCollector::collectArcTo(unsigned /* id */, unsigned le
     end.insert("svg:x", m_scale*m_x);
     end.insert("svg:y", m_scale*m_y);
     end.insert("libwpg:path-action", "L");
-    m_currentGeometry.push_back(end);
+    if (!m_noFill && !m_noShow)
+      m_currentFillGeometry.push_back(end);
+    if (!m_noLine && !m_noShow)
+      m_currentLineGeometry.push_back(end);
   }
   else
   {
@@ -1374,7 +1410,10 @@ void libvisio::VSDXContentCollector::collectArcTo(unsigned /* id */, unsigned le
     arc.insert("svg:x", m_scale*m_x);
     arc.insert("svg:y", m_scale*m_y);
     arc.insert("libwpg:path-action", "A");
-    m_currentGeometry.push_back(arc);
+    if (!m_noFill && !m_noShow)
+      m_currentFillGeometry.push_back(arc);
+    if (!m_noLine && !m_noShow)
+      m_currentLineGeometry.push_back(arc);
   }
 }
 
@@ -1430,7 +1469,10 @@ void libvisio::VSDXContentCollector::collectNURBSTo(unsigned /* id */, unsigned
     transformPoint(nextX, nextY);
     NURBS.insert("svg:x", m_scale*nextX);
     NURBS.insert("svg:y", m_scale*nextY);
-    m_currentGeometry.push_back(NURBS);
+    if (!m_noFill && !m_noShow)
+      m_currentFillGeometry.push_back(NURBS);
+    if (!m_noLine && !m_noShow)
+      m_currentLineGeometry.push_back(NURBS);
   }
 
   m_originalX = x2;
@@ -1442,8 +1484,10 @@ void libvisio::VSDXContentCollector::collectNURBSTo(unsigned /* id */, unsigned
   NURBS.insert("libwpg:path-action", "L");
   NURBS.insert("svg:x", m_scale*m_x);
   NURBS.insert("svg:y", m_scale*m_y);
-  m_currentGeometry.push_back(NURBS);
-
+  if (!m_noFill && !m_noShow)
+    m_currentFillGeometry.push_back(NURBS);
+  if (!m_noLine && !m_noShow)
+    m_currentLineGeometry.push_back(NURBS);
 }
 
 double libvisio::VSDXContentCollector::_NURBSBasis(unsigned knot, unsigned degree, double point, const std::vector<double> &knotVector)
@@ -1471,6 +1515,7 @@ double libvisio::VSDXContentCollector::_NURBSBasis(unsigned knot, unsigned degre
 void libvisio::VSDXContentCollector::collectNURBSTo(unsigned id, unsigned level, double x2, double y2, double knot, double knotPrev, double weight, double weightPrev, unsigned dataID)
 {
   std::map<unsigned, NURBSData>::const_iterator iter;
+  std::map<unsigned, NURBSData>::const_iterator iterEnd;
   NURBSData data;
   if (dataID == 0xFFFFFFFE) // Use stencil NURBS data
   {
@@ -1490,13 +1535,15 @@ void libvisio::VSDXContentCollector::collectNURBSTo(unsigned id, unsigned level,
     }
     dataID = tmpElement->m_dataID;
     iter = m_stencilShape->m_nurbsData.find(dataID);
+    iterEnd =  m_stencilShape->m_nurbsData.end();
   }
   else // No stencils involved, directly get dataID and fill in missing parts
   {
     iter = m_NURBSData.find(dataID);
+    iterEnd = m_NURBSData.end();
   }
 
-  if (iter != m_NURBSData.end())
+  if (iter != iterEnd)
   {
     data = iter->second;
     data.knots.push_back(knot);
@@ -1527,7 +1574,10 @@ void libvisio::VSDXContentCollector::collectPolylineTo(unsigned /* id */ , unsig
     polyline.insert("libwpg:path-action", "L");
     polyline.insert("svg:x", m_scale*points[i].first);
     polyline.insert("svg:y", m_scale*points[i].second);
-    m_currentGeometry.push_back(polyline);
+    if (!m_noFill && !m_noShow)
+      m_currentFillGeometry.push_back(polyline);
+    if (!m_noLine && !m_noShow)
+      m_currentLineGeometry.push_back(polyline);
   }
 
   m_originalX = x;
@@ -1538,7 +1588,10 @@ void libvisio::VSDXContentCollector::collectPolylineTo(unsigned /* id */ , unsig
   polyline.insert("libwpg:path-action", "L");
   polyline.insert("svg:x", m_scale*m_x);
   polyline.insert("svg:y", m_scale*m_y);
-  m_currentGeometry.push_back(polyline);
+  if (!m_noFill && !m_noShow)
+    m_currentFillGeometry.push_back(polyline);
+  if (!m_noLine && !m_noShow)
+    m_currentLineGeometry.push_back(polyline);
 }
 
 /* Polyline with incomplete data */
@@ -1771,6 +1824,7 @@ void libvisio::VSDXContentCollector::collectShape(unsigned id, unsigned level, u
   m_noLine = false;
   m_noFill = false;
   m_noShow = false;
+  m_isFirstGeometry = true;
 
   // Save line colour and pattern, fill type and pattern
   m_fillType = "none";
@@ -2169,7 +2223,7 @@ void libvisio::VSDXContentCollector::_handleLevelChange(unsigned level)
         m_NURBSData = m_stencilShape->m_nurbsData;
         m_polylineData = m_stencilShape->m_polylineData;
 
-        if (m_currentGeometry.empty())
+        if (m_currentFillGeometry.empty() && m_currentLineGeometry.empty() && !m_noShow)
         {
           for (unsigned i = 0; i < m_stencilShape->m_geometries.size(); i++)
           {
@@ -2241,8 +2295,9 @@ void libvisio::VSDXContentCollector::endPage()
   }
 }
 
-void libvisio::VSDXContentCollector::endPages()
+void libvisio::VSDXContentCollector::endPages(const std::vector<unsigned> &pageOrder)
 {
+  m_pages.setOrder(pageOrder);
   m_pages.draw(m_painter);
 }
 
