@@ -64,6 +64,11 @@ void libvisio::VSDXStylesCollector::collectForeignData(unsigned /* id */, unsign
   _handleLevelChange(level);
 }
 
+void libvisio::VSDXStylesCollector::collectOLEData(unsigned /* id */, unsigned level, const WPXBinaryData & /* oleData */)
+{
+  _handleLevelChange(level);
+}
+
 void libvisio::VSDXStylesCollector::collectEllipse(unsigned /* id */, unsigned level, double /* cx */, double /* cy */,
     double /* xleft */, double /* yleft */, double /* xtop */, double /* ytop */)
 {
@@ -189,7 +194,8 @@ void libvisio::VSDXStylesCollector::collectShapeId(unsigned /* id */, unsigned l
   m_shapeList.push_back(shapeId);
 }
 
-void libvisio::VSDXStylesCollector::collectForeignDataType(unsigned /* id */, unsigned level, unsigned /* foreignType */, unsigned /* foreignFormat */)
+void libvisio::VSDXStylesCollector::collectForeignDataType(unsigned /* id */, unsigned level, unsigned /* foreignType */, unsigned /* foreignFormat */,
+    double /* offsetX */, double /* offsetY */, double /* width */, double /* height */)
 {
   _handleLevelChange(level);
 }
@@ -232,7 +238,7 @@ void libvisio::VSDXStylesCollector::collectText(unsigned /*id*/, unsigned level,
 }
 
 void libvisio::VSDXStylesCollector::collectVSDXParaStyle(unsigned /* id */ , unsigned level, unsigned /* charCount */, double /* indFirst */, double /* indLeft */,
-    double /* indRight */, double /* spLine */, double /* spBefore */, double /* spAfter */, unsigned char /* align */)
+    double /* indRight */, double /* spLine */, double /* spBefore */, double /* spAfter */, unsigned char /* align */, unsigned /* flags */)
 {
   _handleLevelChange(level);
 }
@@ -310,11 +316,11 @@ void libvisio::VSDXStylesCollector::collectFillStyle(unsigned /*id*/, unsigned l
 }
 
 void libvisio::VSDXStylesCollector::collectParaIXStyle(unsigned /*id*/, unsigned level, unsigned charCount, double indFirst, double indLeft, double indRight,
-    double spLine, double spBefore, double spAfter, unsigned char align)
+    double spLine, double spBefore, double spAfter, unsigned char align, unsigned flags)
 {
   _handleLevelChange(level);
   if (!m_paraStyle)
-    m_paraStyle = new VSDXParaStyle(charCount, indFirst, indLeft, indRight, spLine, spBefore, spAfter, align);
+    m_paraStyle = new VSDXParaStyle(charCount, indFirst, indLeft, indRight, spLine, spBefore, spAfter, align, flags);
 }
 
 
@@ -372,19 +378,19 @@ void libvisio::VSDXStylesCollector::endPage()
   _handleLevelChange(0);
   m_groupXFormsSequence.push_back(m_groupXForms);
   m_groupMembershipsSequence.push_back(m_groupMemberships);
-  while (m_groupShapeOrder.size())
+  while (!m_groupShapeOrder.empty())
   {
     for (std::list<unsigned>::iterator j = m_pageShapeOrder.begin(); j != m_pageShapeOrder.end();)
     {
       std::map<unsigned, std::list<unsigned> >::iterator iter = m_groupShapeOrder.find(*j);
       if (m_groupShapeOrder.end() != iter)
       {
-        j++;
+        ++j;
         m_pageShapeOrder.splice(j, iter->second, iter->second.begin(), iter->second.end());
         m_groupShapeOrder.erase(iter);
       }
       else
-        j++;
+        ++j;
     }
   }
   m_documentPageShapeOrders.push_back(m_pageShapeOrder);
@@ -435,7 +441,7 @@ void libvisio::VSDXStylesCollector::_handleLevelChange(unsigned level)
 
 void libvisio::VSDXStylesCollector::_flushShapeList()
 {
-  if (!m_shapeList.size())
+  if (m_shapeList.empty())
     return;
 
   if (m_isShapeStarted)
